from datetime import timedelta
import os
import time
from pathlib import Path
import psutil
from typing import Dict, Union, Tuple, Optional
import random

import numpy as np
import torch
import gpytorch
import pickle
import matplotlib.pyplot as plt
from gpytorch.mlls import SumMarginalLogLikelihood
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.models import IndependentModelList
from gpytorch.likelihoods import LikelihoodList
from sklearn.preprocessing import StandardScaler

from gpforecaster import __version__
from gpforecaster.utils.logger import Logger
from gpforecaster.model.gp import ExactGPModel
from gpforecaster.model.mean_functions import PiecewiseLinearMean
from gpforecaster.results.calculate_metrics import CalculateResultsBottomUp


class GPF:
    """
    A class for training and forecasting hierarchical time series datasets with Gaussian Process models.

    Parameters:
        dataset: Name of the dataset.
        groups: Dictionary containing the train and predict groups.
        input_dir: Directory where input files are stored
        n_samples: Number of samples to draw from the posterior distribution
        store_prediction_samples: Whether to store the prediction samples
        store_prediction_points: Whether to store the prediction points.
        log_dir: Directory where log files are stored. Default is '.'.
        perc_inducing_points: Percentage of inducing points to use.

    Attributes:
        dataset (str): Name of the dataset.
        groups (dict): Dictionary containing the train and predict groups.
        input_dir (str): Directory where input files are stored.
        timer_start (float): Start time of the training process.
        wall_time_train (float): Wall time for training.
        wall_time_predict (float): Wall time for prediction.
        wall_time_total (float): Total wall time.
        n_samples (int): Number of samples to draw from the posterior distribution.
        store_prediction_samples (bool): Whether to store the prediction samples.
        store_prediction_points (bool): Whether to store the prediction points.
        train_x (torch.Tensor): Tensor with the training data input.
        train_y (torch.Tensor): Tensor with the training data output.
        test_x (torch.Tensor): Tensor with the test data input.
        test_y (torch.Tensor): Tensor with the test data output.
        losses (list): List of losses during training.
        val_losses (list): List of validation losses during training.
        model_version (str): Version of the model.
        logger_train (Logger): Logger for the training process.
        logger_metrics (Logger): Logger for the metrics.
    """

    def __init__(
        self,
        dataset: str,
        groups: dict,
        input_dir: str = "./",
        n_samples: int = 500,
        store_prediction_samples: bool = False,
        store_prediction_points: bool = False,
        log_dir: str = ".",
        perc_inducing_points: float = 0.5,
    ):
        self.dataset = dataset
        self.groups = groups
        self.input_dir = input_dir
        self.timer_start = time.time()
        self.wall_time_train = None
        self.wall_time_predict = None
        self.wall_time_total = None
        self.groups, self.scaler = self._preprocess(self.groups)
        self._create_directories()
        self.n_samples = n_samples
        self.store_prediction_samples = store_prediction_samples
        self.store_prediction_points = store_prediction_points

        self.train_x = torch.arange(groups["train"]["n"])
        self.train_x = self.train_x.type(torch.DoubleTensor)
        self.train_x = self.train_x.unsqueeze(-1)
        self.train_y = torch.from_numpy(groups["train"]["data"])

        self.test_x = torch.arange(groups["train"]["n"], groups["predict"]["n"])
        self.test_x = self.test_x.type(torch.DoubleTensor)
        self.test_x = self.test_x.unsqueeze(-1)
        self.test_y = torch.from_numpy(
            groups["predict"]["data_matrix"][groups["train"]["n"] :]
        )

        self.losses = []
        self.val_losses = []
        self.mll = None

        self.model_version = __version__

        self.logger_train = Logger(
            "train", dataset=self.dataset, to_file=True, log_dir=log_dir
        )
        self.logger_metrics = Logger(
            "metrics", dataset=self.dataset, to_file=True, log_dir=log_dir
        )

    def _create_directories(self):
        # Create directory to store results if does not exist
        Path(f"{self.input_dir}results").mkdir(parents=True, exist_ok=True)

    def _preprocess(self, groups):
        scaler = StandardScaler()
        scaler.fit(self.groups["train"]["data"])
        groups["train"]["data"] = scaler.transform(groups["train"]["data"])
        groups["predict"]["data_matrix"] = scaler.transform(
            groups["predict"]["data_matrix"]
        )

        return groups, scaler

    def _build_mixtures(self):
        """
        Build the mixtures matrix.

        Returns:
            known_mixtures (np.ndarray): The mixtures matrix.
            n_groups (int): number of groups.

        Example:
                Group1     |   Group2
            GP1, GP2, GP3  | GP1, GP2
            0  , 1  , 1    | 0  , 1
            1  , 0  , 0    | 1  , 0
            0  , 1, , 1    | 0  , 1
            1  , 0  , 1    | 1  , 0
        """
        idxs = []
        for k, val in self.groups["train"]["groups_idx"].items():
            idxs.append(val)

        idxs_t = np.array(idxs).T

        n_groups = np.sum(
            np.fromiter(self.groups["train"]["groups_n"].values(), dtype="int32")
        )
        known_mixtures = np.zeros((self.groups["train"]["s"], n_groups))
        k = 0
        for j in range(self.groups["train"]["g_number"]):
            for i in range(np.max(idxs_t[:, j]) + 1):
                idx_to_1 = np.where(idxs_t[:, j] == i)
                known_mixtures[:, k][idx_to_1] = 1
                k += 1

        top_level = np.ones((known_mixtures.shape[0], 1))
        known_mixtures = np.concatenate((known_mixtures, top_level), axis=1)
        n_groups += 1

        return known_mixtures, n_groups

    def _build_cov_matrices(self):
        """
        Build the covariance matrices.

        Parameters:
            mixtures (np.ndarray): The mixtures matrix.

        Returns:
            covs (list): List of covariance functions.
            known_mixtures (np.ndarray): The mixtures matrix.
            n_groups (int): number of groups.
        """
        known_mixtures, n_groups = self._build_mixtures()
        covs = []
        for i in range(1, n_groups + 1):
            # RBF kernel
            rbf_kernel = gpytorch.kernels.RBFKernel()
            rbf_kernel.lengthscale = torch.tensor([1.0])
            scale_rbf_kernel = gpytorch.kernels.ScaleKernel(rbf_kernel)
            scale_rbf_kernel.outputscale = torch.tensor([0.5])

            # Periodic Kernel
            periodic_kernel = gpytorch.kernels.PeriodicKernel()
            periodic_kernel.period_length = torch.tensor([self.groups["seasonality"]])
            periodic_kernel.lengthscale = torch.tensor([0.5])
            scale_periodic_kernel = gpytorch.kernels.ScaleKernel(periodic_kernel)
            scale_periodic_kernel.outputscale = torch.tensor([1.5])

            # Cov Matrix
            cov = scale_rbf_kernel + scale_periodic_kernel
            covs.append(cov)

        return covs, known_mixtures, n_groups

    def _apply_mixture_cov_matrices(self):
        """
        Apply the mixture covariance matrices and create the final list of covariance functions.

        Returns:
            mixed_covs (cov): The list of mixture covariance functions.
        """
        covs, known_mixtures, n_groups = self._build_cov_matrices()

        # apply mixtures to covariances
        selected_covs = []
        mixed_covs = []
        for i in range(self.groups["train"]["s"]):
            mixture_weights = known_mixtures[i]
            for w_ix in range(n_groups):
                w = mixture_weights[w_ix]
                if w == 1.0:
                    selected_covs.append(covs[w_ix])
            mixed_cov = selected_covs[0]
            for cov in range(1, len(selected_covs)):
                mixed_cov += selected_covs[
                    cov
                ]  # because GP(cov1 + cov2) = GP(cov1) + GP(cov2)
            mixed_covs.append(mixed_cov)
            selected_covs = []  # clear out cov list

        return mixed_covs

    def _build_model(
        self, x: torch.Tensor, y: torch.Tensor
    ) -> Tuple[list[GaussianLikelihood], list[ExactGPModel]]:
        """
        Build the model.

        Parameters:
            x: Measures
            y: Observations.

        Returns:
            likelihood_list: List of GP models.
            model_list: List of likelihoods.
        """
        mixed_covs = self._apply_mixture_cov_matrices()
        n_changepoints = 4
        changepoints = np.linspace(0, self.groups["train"]["n"], n_changepoints + 2)[
            1:-1
        ]

        model_list = []
        likelihood_list = []
        for i in range(self.groups["train"]["s"]):
            likelihood_list.append(GaussianLikelihood())
            model_list.append(
                ExactGPModel(
                    x,
                    y[:, i],
                    likelihood_list[i],
                    mixed_covs[i],
                    changepoints,
                    PiecewiseLinearMean,
                )
            )

        return likelihood_list, model_list

    def early_stopping(self, patience: int):
        losses = [x for x in self.val_losses if x is not None]
        losses.reverse()
        non_decreasing = 0
        for x, y in zip(losses, losses[1:]):
            if np.round(x, 2) >= np.round(y, 2):
                non_decreasing += 1
            else:
                break

        if non_decreasing > patience:
            return True
        else:
            return False

    @staticmethod
    def _create_scheduler(
        scheduler_type: str, optimizer: torch.optim.Optimizer, epochs: int, gamma: float
    ) -> Optional[torch.optim.lr_scheduler._LRScheduler]:
        """
        Creates a scheduler for the learning rate of an optimizer.

        Parameters
            scheduler_type: The type of scheduler to use. One of 'step', 'exponential', 'cosine', or 'none'.
            optimizer: The optimizer for which to create the scheduler.
            epochs: The number of epochs to train the model.

        Returns
            Optional
                The created scheduler, or None if `scheduler_type` is 'none'.
        """
        if scheduler_type == "step":
            scheduler = torch.optim.lr_scheduler.StepLR(
                optimizer, step_size=10, gamma=gamma
            )
        elif scheduler_type == "exponential":
            scheduler = torch.optim.lr_scheduler.ExponentialLR(optimizer, gamma=gamma)
        elif scheduler_type == "cosine":
            scheduler = torch.optim.lr_scheduler.CosineAnnealingLR(
                optimizer, T_max=epochs
            )
        else:
            scheduler = None
        return scheduler

    def train(
        self,
        lr: float = 1e-2,
        epochs: int = 500,
        early_stopping: bool = True,
        patience: int = 4,
        weight_decay: float = 1e-4,
        verbose: bool = True,
        track_mem: bool = False,
        scheduler_type: str = "cosine",
        gamma_rate: float = 0.95,
    ) -> Tuple[IndependentModelList, LikelihoodList]:
        """
        Train the model.

        Parameters:
            lr: Learning rate.
            epochs: Number of epochs
            early_stopping: Perform early stopping
            patience: Parameter to early stopping
            weight_decay: Weight decay
            verbose: Print outputs when training
            track_mem: Track and log RAM usage
            scheduler_type: Type of scheduler to use to update the learning rate
            gamma_rate: Gamma rate for the scheduler types

        Returns:
            tuple: Tuple containing the trained model and the likelihood.
        """
        i = 0
        loss = None
        val_loss = None

        likelihood_list, model_list = self._build_model(x=self.train_x, y=self.train_y)

        model = IndependentModelList(*model_list)
        likelihood = LikelihoodList(*likelihood_list)

        self.mll = SumMarginalLogLikelihood(likelihood, model)

        model.train()
        likelihood.train()

        optimizer = torch.optim.Adam(
            model.parameters(), lr=lr, weight_decay=weight_decay
        )  # Includes GaussianLikelihood parameters
        scheduler = self._create_scheduler(
            scheduler_type, optimizer, epochs, gamma_rate
        )

        for i in range(epochs):
            optimizer.zero_grad()
            output = model(*model.train_inputs)
            loss = -self.mll(output, model.train_targets)
            self.losses.append(loss.item())

            if early_stopping:
                if i % 5 == 0:
                    model.eval()
                    likelihood.eval()
                    val_loss = self.validate(model)
                    if self.early_stopping(patience=patience):
                        break
                    if verbose:
                        print(
                            f"Iter {i}/{epochs} - Loss: {np.round(loss.item(), 3)}, Validation Loss: {np.round(val_loss, 3)}"
                        )
                    # switch to train mode
                    model.train()
                    likelihood.train()
                else:
                    self.val_losses.append(None)
            else:
                if verbose:
                    print(f"Iter {i}/{epochs} - Loss: {np.round(loss.item(), 3)}")

            loss.backward()
            optimizer.step()
            # Step the scheduler
            if scheduler:
                scheduler.step()

            if i % 30 == 0 and track_mem:
                # Track RAM usage
                process = psutil.Process(os.getpid())
                mem = process.memory_info().rss / (1024**3)
                self.logger_train.info(f"train used {mem:.3f} GB of RAM")

        self.wall_time_train = time.time() - self.timer_start
        td = timedelta(seconds=int(time.time() - self.timer_start))
        self.logger_train.info(f"Num epochs {i}")
        self.logger_train.info(f"wall time train {str(td)}")

        self.logger_train.info(f"Loss {np.round(loss.item(), 2)}")
        if early_stopping:
            self.logger_train.info(f"Val Loss {np.round(val_loss, 2)}")

        return model, likelihood

    def validate(self, model: IndependentModelList) -> float:
        """
        Validate the model.

        Parameters:
            model (ExactGPModel): The GP model.
            mll ():

        Returns:
            float: The negative log likelihood of the model on the validation set.
        """
        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            likelihood_list_val, model_list_val = self._build_model(
                x=self.test_x, y=self.test_y
            )
            model_val = gpytorch.models.IndependentModelList(*model_list_val)
            val_output = model(*model_val.train_inputs)
            val_loss = -self.mll(val_output, model_val.train_targets)
            self.val_losses.append(float(val_loss.item()))
        return val_loss.item()

    def plot_losses(self, start_idx: int):
        n_iterations = np.arange(len(self.losses[start_idx:]))
        plt.plot(n_iterations, self.losses[start_idx:])
        plt.plot(n_iterations, self.val_losses[start_idx:], marker="*")
        timestamp = time.strftime("%Y%m%d-%H%M%S", time.gmtime())
        plt.savefig(
            f"./plots/gpf_loss_{self.dataset}_{timestamp}.pdf",
            format="pdf",
            bbox_inches="tight",
        )
        plt.show()

    def predict(
        self,
        model: IndependentModelList,
        likelihood: LikelihoodList,
        clip: bool = True,
    ) -> np.ndarray:
        """
        Make predictions with the model.

        Parameters:
            model: The GP model.
            likelihood: The likelihood function.
            clip: Whether to clip negative predictions to zero. Default is True.

        Returns:
            numpy.ndarray: Array of shape (n_samples, n_prediction_points, n_groups)
                containing the prediction samples.
        """
        timer_start = time.time()

        model.eval()
        likelihood.eval()

        with torch.no_grad(), gpytorch.settings.fast_pred_var():
            test_x = torch.arange(self.groups["predict"]["n"]).type(torch.DoubleTensor)
            predictions = likelihood(
                *model(*[test_x for i in range(self.groups["predict"]["s"])])
            )

        i = 0
        samples = np.zeros(
            (self.n_samples, self.groups["predict"]["n"], self.groups["predict"]["s"])
        )
        for pred in predictions:
            samples[:, :, i] = np.random.normal(
                pred.mean.detach().numpy(),
                np.sqrt(pred.variance.detach().numpy()),
                size=(self.n_samples, self.groups["predict"]["n"]),
            )
            i += 1

        samples = np.transpose(samples, (1, 2, 0))

        # transform back the data
        samples = (
            samples * np.sqrt(self.scaler.var_)[np.newaxis, :, np.newaxis]
        ) + self.scaler.mean_[np.newaxis, :, np.newaxis]
        self.groups["train"]["data"] = self.scaler.inverse_transform(
            self.groups["train"]["data"]
        )
        self.groups["predict"]["data_matrix"] = self.scaler.inverse_transform(
            self.groups["predict"]["data_matrix"]
        )

        # Clip predictions to 0 if there are negative numbers
        if clip:
            samples[samples < 0] = 0

        self.wall_time_predict = time.time() - timer_start
        return samples

    def store_metrics(self, res):
        with open(
            f"{self.input_dir}results/results_gp_cov_{self.dataset}_{self.model_version}.pickle",
            "wb",
        ) as handle:
            pickle.dump(res, handle, pickle.HIGHEST_PROTOCOL)

    def metrics(
        self, samples: np.ndarray
    ) -> Dict[str, Dict[str, Union[float, np.ndarray]]]:
        """
        Calculate evaluation metrics for the predictions.

        Parameters:
            samples: Array of shape (n_samples, n_prediction_points, n_groups)
                containing the prediction samples.

        Returns:
            dict: Dictionary with the evaluation metrics. The keys are the metric names,
                and the values are dictionaries with the results for each group.
        """
        calc_results = CalculateResultsBottomUp(
            samples,
            self.groups,
            self.store_prediction_samples,
            self.store_prediction_points,
        )
        res = calc_results.calculate_metrics()
        for metric, results in res.items():
            for group, result in results.items():
                if "ind" not in group:
                    self.logger_metrics.info(
                        f"{metric} for {group}: {np.round(result, 2)}"
                    )

        self.wall_time_total = time.time() - self.timer_start

        res["wall_time"] = {}
        res["wall_time"]["wall_time_train"] = self.wall_time_train
        res["wall_time"]["wall_time_predict"] = self.wall_time_predict
        res["wall_time"]["wall_time_total"] = self.wall_time_total

        return res



import pickle
import shutil
import unittest

import numpy as np
from numpy.testing import assert_array_almost_equal
import tsaugmentation as tsag

from gpforecaster.model.gpf import GPF


class TestModel(unittest.TestCase):
    def setUp(self):
        self.data = tsag.preprocessing.PreprocessDatasets("prison").apply_preprocess()
        self.n = self.data["predict"]["n"]
        self.s = self.data["train"]["s"]
        shutil.rmtree("./data/original_datasets")
        self.gpf = GPF("prison", self.data)

    def test_correct_train(self):
        model, like = self.gpf.train(epochs=10)
        self.assertIsNotNone(model)

    def test_predict_shape(self):
        model, like = self.gpf.train(epochs=10)
        samples = self.gpf.predict(model, like)
        self.assertTrue(samples.shape == (self.n, self.s, 500))

    def test_results_interval(self):
        model, like = self.gpf.train(epochs=100)
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)
        self.assertLess(res["mase"]["bottom"], 2.5)

    def test_wall_time(self):
        model, like = self.gpf.train(epochs=10)
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)
        self.assertLess(res["wall_time"]["wall_time_total"], 100)

    def test_output_results(self):
        model, like = self.gpf.train(epochs=10)
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)
        self.gpf.store_metrics(res)
        with open(
            f"{self.gpf.input_dir}results/results_gp_cov_{self.gpf.dataset}_{self.gpf.model_version}.pickle",
            "rb",
        ) as handle:
            output_res = pickle.load(file=handle)
        self.assertIsNotNone(output_res["mase"]["bottom"])



import unittest
import tsaugmentation as tsag
from gpforecaster.model.gpf import GPF
import shutil


class TestModel(unittest.TestCase):

    def setUp(self):
        self.preproc = tsag.preprocessing.PreprocessDatasets('tourism', test_size=228*10)
        self.data = self.preproc._tourism()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        self.gpf = GPF(dataset='tourism', groups=self.data,
                       store_prediction_samples=True,
                       store_prediction_points=True)

    def test_results_mean_and_prediction_interval(self):
        model, like = self.gpf.train(epochs=10)
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)

        # Test shape of results
        self.assertTrue(res['mase']['bottom_ind'].shape == (self.s, ))
        self.assertTrue(res['CRPS']['bottom_ind'].shape == (self.s, ))
        self.assertTrue(res['rmse']['bottom_ind'].shape == (self.s, ))

        # Test shape of predictions
        # Number of prediction samples for the bottom time series (n_points, n_series, n_samples)
        self.assertTrue(res['predictions']['samples']['bottom'].shape == (self.n, self.s, 500))
        self.assertTrue(res['predictions']['points']['bottom'].shape == (self.data['h'], self.s))
        # Number of prediction samples for the total time series
        self.assertTrue(res['predictions']['samples']['total'].shape == (self.n, 500))
        self.assertTrue(res['predictions']['points']['total'].shape == (self.data['h'],))
        # Test number of objects predicted and stored
        self.assertTrue(len(res['predictions']['samples']) == 12)
        self.assertTrue(len(res['predictions']['points']) == 12)

    def test_results_mean_and_prediction_interval_without_storing_results(self):
        self.gpf = GPF(dataset='tourism', groups=self.data,
                       store_prediction_samples=False,
                       store_prediction_points=False)
        model, like = self.gpf.train(epochs=10)
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)

        # Test shape of results
        self.assertTrue(res['mase']['bottom_ind'].shape == (self.s, ))
        self.assertTrue(res['CRPS']['bottom_ind'].shape == (self.s, ))
        self.assertTrue(res['rmse']['bottom_ind'].shape == (self.s, ))

        # Test shape of predictions
        # Test number of objects predicted and stored
        self.assertTrue(len(res) == 4)




from typing import Tuple, Dict
import random

from gpforecaster.model.gpf import GPF
from gpforecaster.utils.logger import Logger


def optimize_hyperparameters(
    dataset_name: str,
    hierarchical_data: Dict,
    num_trials: int = 20,
    learning_rates: Tuple[float] = (1e-2, 1e-3),
    weight_decays: Tuple[float] = (1e-3, 1e-4, 1e-5),
    scheduler_types: Tuple[str] = ("step", "exponential", "cosine", "none"),
    gamma_rates: Tuple[int] = (0.1, 0.9, 0.95, 0.8),
) -> Tuple[float, float, str, float]:
    """
    Performs a random search to optimize the hyperparameters of the model.

    Parameters
        dataset: Dataset to run hyperparameter tuning
        hierarchical_data: Dict containing data and metadata
        num_trials: Number of trials to perform

    Returns
        best_hyperparameters : A tuple containing the optimal learning rate,
            weight decay, and scheduler type.
    """
    logger_tuning = Logger("hyperparameter_tuning", dataset=dataset_name, to_file=True)

    results = []

    for trial in range(num_trials):
        # Sample hyperparameters randomly
        learning_rate = random.choice(learning_rates)
        weight_decay = random.choice(weight_decays)
        scheduler_type = random.choice(scheduler_types)
        gamma_rate = random.choice(gamma_rates)

        gpf = GPF(dataset=dataset_name, groups=hierarchical_data)

        # Evaluate the performance with the sampled hyperparameters
        model, _ = gpf.train(
            lr=learning_rate,
            weight_decay=weight_decay,
            scheduler_type=scheduler_type,
            gamma_rate=gamma_rate,
        )

        val_loss = gpf.validate(model)

        results.append((learning_rate, weight_decay, scheduler_type, gamma_rate, val_loss))

    results.sort(key=lambda x: x[4])
    best_hyperparameters = results[0]

    logger_tuning.info(
        f"Best hyperparameters: learning rate = {best_hyperparameters[0]}, weight decay = {best_hyperparameters[1]}, scheduler = {best_hyperparameters[2]}, gamma = {best_hyperparameters[3]},"
    )
    logger_tuning.info(f"Validation loss: {best_hyperparameters[4]}")

    return best_hyperparameters



import unittest
import tsaugmentation as tsag
from gpforecaster.model.gpf import GPF
import shutil


class TestModel(unittest.TestCase):

    def setUp(self):
        self.data = tsag.preprocessing.PreprocessDatasets('prison').apply_preprocess()
        self.n = self.data['predict']['n']
        self.s = self.data['train']['s']
        shutil.rmtree("./data/original_datasets")
        self.gpf = GPF('prison', self.data)

    def test_calculate_metrics_dict(self):
        model, like = self.gpf.train(epochs=100)
        samples = self.gpf.predict(model, like)
        res = self.gpf.metrics(samples)
        self.assertLess(res['mase']['bottom'], 2.5)



import unittest
from gpforecaster.model.gpf import GPF
import tsaugmentation as tsag
import timeit


class TestModel(unittest.TestCase):
    def setUp(self):
        self.data = tsag.preprocessing.PreprocessDatasets("prison").apply_preprocess()
        self.n = self.data["predict"]["n"]
        self.s = self.data["train"]["s"]
        self.gpf = GPF("prison", self.data, log_dir="..")

    def test_early_stopping_fn(self):
        self.gpf.val_losses = [5.1, 5.2, 4.9, 5.0, 5.1, 5.2]
        res = self.gpf.early_stopping(2)
        self.assertTrue(res)

    def test_early_stopping_w_patience(self):
        model, like = self.gpf.train(epochs=200, patience=4, track_mem=True)
        self.gpf.plot_losses(5)
        self.assertEqual(len(self.gpf.losses), 151)

    def test_compare_execution_times(self):
        n_iter = 100
        elapsed_time_es = timeit.timeit(
            lambda: self.gpf.train(epochs=n_iter, verbose=False), number=1
        )
        print(f"Elapsed time with Early Stopping: {elapsed_time_es} seconds")
        elapsed_time = timeit.timeit(
            lambda: self.gpf.train(epochs=n_iter, early_stopping=False, verbose=False),
            number=1,
        )
        print(f"Elapsed time: {elapsed_time} seconds")

    def test_without_early_stopping(self):
        model, like = self.gpf.train(epochs=500, early_stopping=False, track_mem=True)
        self.assertEqual(len(self.gpf.losses), 500)



import gpytorch


class ExactGPModel(gpytorch.models.ExactGP):
    def __init__(self, train_x, train_y, likelihood, cov, changepoints, mean_func):
        super().__init__(train_x, train_y, likelihood)
        self.changepoints = changepoints
        self.mean_module = mean_func(self.changepoints)
        self.covar_module = cov

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)


class SparseGPModel(gpytorch.models.ExactGP):
    def __init__(
        self,
        train_x,
        train_y,
        likelihood,
        cov,
        changepoints,
        mean_func,
        inducing_points,
    ):
        super().__init__(train_x, train_y, likelihood)
        self.changepoints = changepoints
        self.mean_module = mean_func(self.changepoints)
        # Use an InducingPointKernel instead of a standard kernel
        self.covar_module = gpytorch.kernels.InducingPointKernel(
            cov, inducing_points=inducing_points, likelihood=likelihood
        )

    def forward(self, x):
        mean_x = self.mean_module(x)
        covar_x = self.covar_module(x)
        return gpytorch.distributions.MultivariateNormal(mean_x, covar_x)



import gpytorch
import torch


class LinearMean(gpytorch.means.Mean):
    def __init__(self, input_size, batch_shape=torch.Size(), bias=True):
        super().__init__()
        self.register_parameter(name="weights", parameter=torch.nn.Parameter(torch.randn(*batch_shape, input_size, 1)))
        if bias:
            self.register_parameter(name="bias", parameter=torch.nn.Parameter(torch.randn(*batch_shape, 1)))
        else:
            self.bias = None

    def forward(self, x):
        x = x.float()
        res = x.matmul(self.weights).squeeze(-1)
        if self.bias is not None:
            res = res + self.bias

        return res


class PiecewiseLinearMean(gpytorch.means.Mean):
    def __init__(self, changepoints):
        super().__init__()
        self.changepoints = changepoints
        self.register_parameter(name="k",
                                parameter=torch.nn.Parameter(torch.tensor([[0.1]])))
        self.register_parameter(name="m",
                                parameter=torch.nn.Parameter(torch.tensor([[0.1]])))
        self.register_parameter(name="b",
                                parameter=torch.nn.Parameter(torch.tile(torch.tensor([0.1]), (len(changepoints),))))

    def forward(self, x):
        x = x.float()
        A = (0.5 * (1.0 + torch.sgn(torch.tile(x.reshape((-1, 1)), (1, 4)) - self.changepoints))).float()

        res = ((self.k + torch.matmul(A, self.b.reshape((-1, 1)))) * x
               + (self.m + torch.matmul(A, (-torch.from_numpy(self.changepoints).float() * self.b))).reshape(-1, 1))

        return res.reshape((-1,))



import numpy as np
from sklearn.metrics import mean_squared_error
import properscoring as ps


class CalculateResultsBase:
    """Calculate the results and store them using pickle files

    Currently we have implemented MASE and RMSE.

    Attributes:
        pred_samples (array): predictions of shape [number of samples, h, number of series]
            - we transform it immediately to the shape [h, number of series] by averaging over the samples
        groups (obj): dict containing the data and several other attributes of the dataset

    """

    def __init__(self, groups, samples=500):
        self.groups = groups
        self.seas = self.groups['seasonality']
        self.h = self.groups['h']
        self.n = self.groups['predict']['n']
        self.s = self.groups['predict']['s']
        self.y_f = self.groups['predict']['data'].reshape(self.s, self.n).T
        self.errs = ['mase', 'rmse', 'CRPS']
        self.levels = list(self.groups['train']['groups_names'].keys())
        self.levels.extend(('bottom', 'total'))
        self.samples = samples

    def mase(self, y, f):
        """Calculates the Mean Absolute Scaled Error

        Args:
            param y (array): original series values shape=[n,s]
            param f (array): predictions shape=[h,s]

        Returns:
            MASE (array): shape=[s,]

        """
        return ((self.n - self.seas) / self.h * (np.sum(np.abs(y[self.n - self.h:self.n, :] - f), axis=0)
                                                 / np.sum(np.abs(y[self.seas:self.n, :] - y[:self.n - self.seas, :]),
                                                          axis=0)))

    def calculate_metrics_for_individual_group(self,
                                               group_name,
                                               y,
                                               predictions_mean,
                                               error_metrics,
                                               predictions_sample=None,
                                               predictions_std=None):
        """Calculates the main metrics for each group

        Args:
            param group_name: group that we want to calculate the error metrics
            param y: original series values with the granularity of the group to calculate
            param predictions_mean: predictions mean with the granularity of the group to calculate
            param error_metrics: dict to add new results
            param predictions_sample: samples of the predictions
            param predictions_variance: variance of the predictions

        Returns:
            error (obj): contains both the error metric for each individual series of each group and the average

        """
        y_p = y[self.n - self.h:self.n, :]
        predictions_sample = predictions_sample[self.n - self.h:self.n, :, :]
        n_s = y_p.shape[1]
        error_metrics['mase'][f'{group_name}_ind'] = np.round(self.mase(y=y,
                                                                        f=predictions_mean), 3)
        error_metrics['mase'][f'{group_name}'] = np.round(np.mean(error_metrics['mase'][f'{group_name}_ind']), 3)
        error_metrics['rmse'][f'{group_name}_ind'] = np.round(mean_squared_error(y_p,
                                                              predictions_mean,
                                                              squared=False,
                                                              multioutput='raw_values'),
                                                              3)
        error_metrics['rmse'][f'{group_name}'] = np.round(np.mean(error_metrics['rmse'][f'{group_name}_ind']), 3)

        if predictions_std is not None:
            error_metrics['CRPS'][f'{group_name}'] = ps.crps_ensemble(y_p,
                                                                      np.random.normal(loc=predictions_mean,
                                                                                       scale=predictions_std,
                                                                                       size=(self.samples,
                                                                                             self.h,
                                                                                             n_s))
                                                                      .reshape((self.h,
                                                                                n_s,
                                                                                -1))).mean()
            error_metrics['CRPS'][f'{group_name}_ind'] = ps.crps_ensemble(y_p,
                                                                          np.random.normal(loc=predictions_mean,
                                                                                           scale=predictions_std,
                                                                                           size=(self.samples,
                                                                                                 self.h,
                                                                                                 n_s))
                                                                          .reshape((self.h,
                                                                                    n_s,
                                                                                    -1))).mean(axis=0)
        else:
            error_metrics['CRPS'][f'{group_name}'] = ps.crps_ensemble(y_p, predictions_sample).mean()
            error_metrics['CRPS'][f'{group_name}_ind'] = ps.crps_ensemble(y_p, predictions_sample).mean(axis=0)

        return error_metrics


class CalculateResultsBottomUp(CalculateResultsBase):
    r"""
    Calculate results for the bottom-up strategy.

    From the prediction of the bottom level series, aggregate the results for the upper levels
    considering the hierarchical structure and compute the error metrics accordingly.

    Parameters
    ----------
    pred_samples : numpy array
        results for the bottom series
    groups : dict
        all the information regarding the different groups
    """

    def __init__(self, pred_samples, groups, store_prediction_samples, store_prediction_points):
        super().__init__(groups=groups)
        self.pred_mean = np.mean(pred_samples, axis=2).reshape(self.n, self.s)[self.n - self.h:self.n, :]
        self.pred_mean_complete = np.mean(pred_samples, axis=2).reshape(self.n, self.s)
        self.pred_samples = pred_samples
        self.store_prediction_samples = store_prediction_samples
        self.store_prediction_points = store_prediction_points

    def compute_error_for_every_group(self, error_metrics):
        """Computes the error metrics for all the groups

        Returns:
            error (obj): - contains all the error metric for each group in the dataset
                         - contains all the predictions for all the groups

        """
        idx_dict_new = dict()
        for group in list(self.groups['predict']['groups_names'].keys()):
            y_g = np.zeros((self.groups['predict']['n'], self.groups['predict']['groups_names'][group].shape[0]))
            f_g = np.zeros((self.h, self.groups['predict']['groups_names'][group].shape[0]))
            s_g = np.zeros((self.n, self.groups['predict']['groups_names'][group].shape[0], self.pred_samples.shape[2]))

            for idx, name in enumerate(self.groups['predict']['groups_names'][group]):
                idx_dict_new[name] = np.where(self.groups['predict']['groups_idx'][group] == idx, 1, 0).reshape((1, -1))

                y_g[:, idx] = np.sum(idx_dict_new[name] * self.y_f, axis=1)
                f_g[:, idx] = np.sum(idx_dict_new[name] * self.pred_mean, axis=1)
                s_g[:, idx, :] = np.sum(idx_dict_new[name][:, :, np.newaxis] * self.pred_samples, axis=1)
                if self.store_prediction_points:
                    error_metrics['predictions']['points'][name] = f_g[:, idx]
                if self.store_prediction_samples:
                    error_metrics['predictions']['samples'][name] = s_g[:, idx, :]

            error_metrics = self.calculate_metrics_for_individual_group(group,
                                                                        y_g,
                                                                        f_g,
                                                                        error_metrics,
                                                                        predictions_sample=s_g)

        return error_metrics

    def bottom_up(self, level, error_metrics):
        """Aggregates the results for all the groups

        Returns:
            error (obj): - contains all the error metric for the specific level

        """
        if level == 'bottom':
            error_metrics = self.calculate_metrics_for_individual_group(level,
                                                                        self.y_f,
                                                                        self.pred_mean,
                                                                        error_metrics,
                                                                        predictions_sample=self.pred_samples)
            if self.store_prediction_points:
                error_metrics['predictions']['points']['bottom'] = self.pred_mean
            if self.store_prediction_samples:
                error_metrics['predictions']['samples']['bottom'] = self.pred_samples
        elif level == 'total':
            error_metrics = self.calculate_metrics_for_individual_group(level,
                                                                        np.sum(self.y_f, axis=1).reshape(-1, 1),
                                                                        np.sum(self.pred_mean, axis=1).reshape(-1, 1),
                                                                        error_metrics,
                                                                        predictions_sample=np.sum(self.pred_samples,
                                                                                                  axis=1)
                                                                        .reshape((self.n, 1, -1)))

            if self.store_prediction_points:
                error_metrics['predictions']['points']['total'] = np.sum(self.pred_mean, axis=1)
            if self.store_prediction_samples:
                error_metrics['predictions']['samples']['total'] = np.sum(self.pred_samples, axis=1)
        elif level == 'groups':
            self.compute_error_for_every_group(error_metrics)

        return error_metrics

    def calculate_metrics(self):
        """Aggregates the results for all the groups

        Returns:
            error (obj): - contains all the error metric for each individual series of each group and the average
                         - contains all the predictions for all the series and groups

        """
        error_metrics = dict()
        error_metrics['mase'] = {}
        error_metrics['rmse'] = {}
        error_metrics['CRPS'] = {}
        if self.store_prediction_points or self.store_prediction_samples:
            error_metrics['predictions'] = {}
        if self.store_prediction_points:
            error_metrics['predictions']['points'] = {}
        if self.store_prediction_samples:
            error_metrics['predictions']['samples'] = {}

        error_metrics = self.bottom_up('bottom', error_metrics)
        error_metrics = self.bottom_up('total', error_metrics)
        error_metrics = self.bottom_up('groups', error_metrics)

        # Aggregate all errors and create the 'all' category
        for err in self.errs:
            error_metrics[err]['all_ind'] = np.squeeze(np.concatenate([error_metrics[err][f'{x}_ind'].reshape((-1, 1))
                                                                       for x in self.levels], 0))
            error_metrics[err]['all'] = np.mean(error_metrics[err]['all_ind'])

        return error_metrics



import unittest
import timeit

import tsaugmentation as tsag

from gpforecaster.model.gpf import GPF
from gpforecaster.model.hyperparameter_tuning import optimize_hyperparameters


class TestModel(unittest.TestCase):
    def setUp(self):
        self.dataset_name = "prison"
        self.data = tsag.preprocessing.PreprocessDatasets(
            self.dataset_name
        ).apply_preprocess()
        self.n = self.data["predict"]["n"]
        self.s = self.data["train"]["s"]

    def test_early_stopping_w_patience(self):
        best_hyperparameters = optimize_hyperparameters(
            dataset_name=self.dataset_name, hierarchical_data=self.data, num_trials=2
        )
        self.assertIsNotNone(best_hyperparameters)
